# Copyright 2020-2023 OpenDR European Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
from filterpy.kalman import KalmanFilter
from opendr.engine.target import BoundingBox3D, TrackingAnnotation3D


class KalmanTracker3D():

    def __init__(
        self,
        boundingBox3D: BoundingBox3D,
        id,
        frame,
        state_dimensions=10,  # x, y, z, rotation_y, l, w, h, speed_x, speed_z, angular_speed
        measurement_dimensions=7,  # x, y, z, rotation_y, l, w, h
        state_transition_matrix=None,
        measurement_function_matrix=None,
        covariance_matrix=None,
        process_uncertainty_matrix=None,
    ):
        super().__init__()

        self.start_frame = frame
        self.last_update_frame = frame
        self.id = id

        self.kalman_filter = KalmanFilter(dim_x=state_dimensions, dim_z=measurement_dimensions)
        self.predictions = []
        self.updates = 0

        if state_transition_matrix is None:
            state_transition_matrix = np.eye(state_dimensions, dtype=np.float32)
            state_transition_matrix[0, -3] = 1
            state_transition_matrix[1, -2] = 1
            state_transition_matrix[2, -1] = 1

        if measurement_function_matrix is None:
            measurement_function_matrix = np.eye(
                measurement_dimensions, state_dimensions, dtype=np.float32
            )

        if covariance_matrix is None:
            covariance_matrix = np.eye(
                state_dimensions, state_dimensions, dtype=np.float32
            ) * 10
            covariance_matrix[7:, 7:] *= 1000

        if process_uncertainty_matrix is None:
            process_uncertainty_matrix = np.eye(
                state_dimensions, state_dimensions, dtype=np.float32
            )
            process_uncertainty_matrix[7:, 7:] *= 0.01

        self.kalman_filter.F = state_transition_matrix
        self.kalman_filter.H = measurement_function_matrix
        self.kalman_filter.P = covariance_matrix
        self.kalman_filter.Q = process_uncertainty_matrix

        location = boundingBox3D.data["location"]
        dimensions = boundingBox3D.data["dimensions"]
        rotation_y = boundingBox3D.data["rotation_y"]

        # [x, y, z, rotation_y, l, w, h]
        self.kalman_filter.x[:measurement_dimensions] = np.array([
            *location, rotation_y, *dimensions
        ]).reshape(-1, 1)

        self.name = boundingBox3D.name
        self.bbox2d = boundingBox3D.bbox2d
        self.action = boundingBox3D.action
        self.alpha = boundingBox3D.alpha
        self.truncated = boundingBox3D.truncated
        self.occluded = boundingBox3D.occluded
        self.confidence = boundingBox3D.confidence

    def update(self, boundingBox3D: BoundingBox3D, frame):

        self.last_update_frame = frame
        self.updates += 1

        location = boundingBox3D.data["location"]
        dimensions = boundingBox3D.data["dimensions"]
        rotation_y = boundingBox3D.data["rotation_y"]

        self.name = boundingBox3D.name
        self.bbox2d = boundingBox3D.bbox2d
        self.action = boundingBox3D.action
        self.alpha = boundingBox3D.alpha
        self.truncated = boundingBox3D.truncated
        self.occluded = boundingBox3D.occluded
        self.confidence = boundingBox3D.confidence

        rotation_y = normalize_angle(rotation_y)
        predicted_rotation_y = self.kalman_filter.x[3]

        if (
            abs(rotation_y - predicted_rotation_y) >= np.pi / 2 and
            abs(rotation_y - predicted_rotation_y) <= np.pi * 1.5
        ):
            predicted_rotation_y = normalize_angle(predicted_rotation_y + np.pi)

        if abs(rotation_y - predicted_rotation_y) >= np.pi * 1.5:
            if rotation_y > 0:
                predicted_rotation_y += np.pi * 2
            else:
                predicted_rotation_y -= np.pi * 2

        self.kalman_filter.x[3] = predicted_rotation_y

        self.kalman_filter.update(np.array([
            *location, rotation_y, *dimensions
        ]))

    def predict(self) -> np.ndarray:
        self.kalman_filter.predict()
        self.kalman_filter.x[3] = normalize_angle(self.kalman_filter.x[3])
        self.predictions.append(self.kalman_filter.x)

        return self.kalman_filter.x

    def tracking_bounding_box_3d(self, frame):
        return TrackingAnnotation3D(
            self.name, self.truncated, self.occluded,
            self.alpha, self.bbox2d,
            self.kalman_filter.x[4:7].reshape(-1),
            self.kalman_filter.x[:3].reshape(-1),
            float(self.kalman_filter.x[3]),
            self.id,
            self.confidence,
            frame,
        )

    def age(self, frame):
        return frame - self.start_frame

    def staleness(self, frame):
        return frame - self.last_update_frame


def normalize_angle(angle):
    if angle >= np.pi:
        angle -= np.pi * 2
    if angle < -np.pi:
        angle += np.pi * 2
    return angle



# Copyright 2020-2023 OpenDR European Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time
from opendr.engine.learners import Learner
from opendr.engine.datasets import DatasetIterator
from opendr.engine.target import BoundingBox3DList
from opendr.perception.object_tracking_3d.ab3dmot.algorithm.ab3dmot import AB3DMOT
from opendr.perception.object_tracking_3d.ab3dmot.algorithm.evaluate import evaluate as evaluate_kitti_tracking
from opendr.perception.object_tracking_3d.ab3dmot.logger import Logger


class ObjectTracking3DAb3dmotLearner(Learner):
    def __init__(
        self,
        device="cpu",
        max_staleness=2,
        min_updates=3,
        state_dimensions=10,  # x, y, z, rotation_y, l, w, h, speed_x, speed_z, angular_speed
        measurement_dimensions=7,  # x, y, z, rotation_y, l, w, h
        state_transition_matrix=None,
        measurement_function_matrix=None,
        covariance_matrix=None,
        process_uncertainty_matrix=None,
        iou_threshold=0.01,
    ):
        # Pass the shared parameters on super's constructor so they can get initialized as class attributes
        super(ObjectTracking3DAb3dmotLearner, self).__init__(
            device=device,
        )

        self.max_staleness = max_staleness
        self.min_updates = min_updates
        self.state_dimensions = state_dimensions
        self.measurement_dimensions = measurement_dimensions
        self.state_transition_matrix = state_transition_matrix
        self.measurement_function_matrix = measurement_function_matrix
        self.covariance_matrix = covariance_matrix
        self.process_uncertainty_matrix = process_uncertainty_matrix
        self.iou_threshold = iou_threshold
        self.model = None

        self.infers_count = 0
        self.infers_time = 0

        self.__create_model()

    def save(self, path):
        raise NotImplementedError("The AB3DMOT Learner does not support saving")

    def load(
        self,
        path,
    ):
        raise NotImplementedError("The AB3DMOT Learner does not support loading")

    def reset(self):
        self.model.reset()

    def fit(
        self,
        dataset,
        val_dataset=None,
        logging_path=None,
        silent=False,
        verbose=False,
    ):
        raise NotImplementedError("The AB3DMOT Learner does not support training")

    def eval(
        self,
        dataset,
        logging_path=None,
        silent=False,
        verbose=False,
        count=None
    ):

        logger = Logger(silent, verbose, logging_path)

        if not isinstance(dataset, DatasetIterator):
            raise ValueError("dataset should be a DatasetIterator")

        if count is None:
            count = len(dataset)

        predictions = []
        ground_truths = []

        for i in range(count):
            self.reset()
            input, ground_truth = dataset[i]
            predictions.append(self.infer(input))
            ground_truths.append(ground_truth)

            logger.log(Logger.LOG_WHEN_NORMAL, "Computing tracklets [" + str(i + 1) + "/" + str(count) + "]", end='\r')

        result = evaluate_kitti_tracking(predictions, ground_truths, log=logger.log)

        logger.close()

        return result

    def infer(self, bounding_boxes_3d_list):

        if self.model is None:
            raise ValueError("No model created")

        is_single_input = True

        if isinstance(bounding_boxes_3d_list, BoundingBox3DList):
            bounding_boxes_3d_list = [bounding_boxes_3d_list]
        elif isinstance(bounding_boxes_3d_list, list):
            is_single_input = False
        else:
            return ValueError(
                "bounding_boxes_3d_list should be a BoundingBox3DList or a list of BoundingBox3DList"
            )

        results = []

        for box_list in bounding_boxes_3d_list:

            t0 = time.time()

            result = self.model.update(box_list)
            results.append(result)

            t0 = time.time() - t0
            self.infers_count += 1
            self.infers_time += t0

        if is_single_input:
            results = results[0]

        return results

    def optimize(self):
        raise Exception("The AB3DMOT Learner does not support optimization")

    def __create_model(self):

        self.model = AB3DMOT(
            frame=0,
            max_staleness=self.max_staleness,
            min_updates=self.min_updates,
            state_dimensions=self.state_dimensions,
            measurement_dimensions=self.measurement_dimensions,
            state_transition_matrix=self.state_transition_matrix,
            measurement_function_matrix=self.measurement_function_matrix,
            covariance_matrix=self.covariance_matrix,
            process_uncertainty_matrix=self.process_uncertainty_matrix,
            iou_threshold=self.iou_threshold,
        )



# Copyright 2020-2023 OpenDR European Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List
import numba
import copy
import numpy as np
from scipy.spatial import ConvexHull


@numba.jit
def polygon_area(x, y):
    return 0.5 * np.abs(np.dot(x, np.roll(y, 1)) - np.dot(y, np.roll(x, 1)))


@numba.jit
def corner_box3d_volume(corners: np.array):  # [8, 3] -> []

    result = (
        np.sqrt(np.sum((corners[0, :] - corners[1, :]) ** 2)) *
        np.sqrt(np.sum((corners[1, :] - corners[2, :]) ** 2)) *
        np.sqrt(np.sum((corners[0, :] - corners[4, :]) ** 2))
    )
    return result


def polygon_clip(subject_polygon, clip_polygon):  # [(x, y)] -> [(x, y)] -> [(x, y))
    def is_inside(p, clip_polygon1, clip_polygon2):
        return (clip_polygon2[0] - clip_polygon1[0]) * (p[1] - clip_polygon1[1]) > (
            clip_polygon2[1] - clip_polygon1[1]
        ) * (p[0] - clip_polygon1[0])

    def intersection(clip_polygon1, clip_polygon2):
        dc = [clip_polygon1[0] - clip_polygon2[0], clip_polygon1[1] - clip_polygon2[1]]
        dp = [s[0] - e[0], s[1] - e[1]]
        n1 = clip_polygon1[0] * clip_polygon2[1] - clip_polygon1[1] * clip_polygon2[0]
        n2 = s[0] * e[1] - s[1] * e[0]
        n3 = 1.0 / (dc[0] * dp[1] - dc[1] * dp[0])
        return [(n1 * dp[0] - n2 * dc[0]) * n3, (n1 * dp[1] - n2 * dc[1]) * n3]

    outputList = subject_polygon
    cp1 = clip_polygon[-1]

    for clip_vertex in clip_polygon:
        cp2 = clip_vertex
        inputList = outputList
        outputList = []
        s = inputList[-1]

        for subjectVertex in inputList:
            e = subjectVertex
            if is_inside(e, cp1, cp2):
                if not is_inside(s, cp1, cp2):
                    outputList.append(intersection(cp1, cp2))
                outputList.append(e)
            elif is_inside(s, cp1, cp2):
                outputList.append(intersection(cp1, cp2))
            s = e
        cp1 = cp2
        if len(outputList) == 0:
            return None
    return outputList


@numba.jit
def convex_hull_intersection(
    polygon1: List[tuple], polygon2: List[tuple]
):  # [(x, y)] -> [(x, y)] -> ([(x, y), []])
    inter_p = polygon_clip(polygon1, polygon2)
    if inter_p is not None:
        hull_inter = ConvexHull(inter_p)
        return inter_p, hull_inter.volume
    else:
        return None, 0.0


def iou3D(corners1, corners2):  # [8, 3] -> [8, 3] -> ([], [])
    # corner points are in counter clockwise order
    rect1 = [(corners1[i, 0], corners1[i, 2]) for i in range(3, -1, -1)]
    rect2 = [(corners2[i, 0], corners2[i, 2]) for i in range(3, -1, -1)]
    area1 = polygon_area(np.array(rect1)[:, 0], np.array(rect1)[:, 1])
    area2 = polygon_area(np.array(rect2)[:, 0], np.array(rect2)[:, 1])
    _, inter_area = convex_hull_intersection(rect1, rect2)
    iou_2d = inter_area / (area1 + area2 - inter_area)
    y_max = min(corners1[0, 1], corners2[0, 1])
    y_min = max(corners1[4, 1], corners2[4, 1])
    inter_vol = inter_area * max(0.0, y_max - y_min)
    vol1 = corner_box3d_volume(corners1)
    vol2 = corner_box3d_volume(corners2)
    iou = inter_vol / (vol1 + vol2 - inter_vol)
    return iou, iou_2d


@numba.jit
def rotation_matrix_y(t):  # [] -> [3, 3]
    c = np.cos(t)
    s = np.sin(t)
    return np.array([[c, 0, s], [0, 1, 0], [-s, 0, c]])


def convert_3dbox_to_8corner(bbox3d_input):  # [7] -> [8, 3]
    bbox3d = copy.copy(bbox3d_input)
    rot_matrix = rotation_matrix_y(bbox3d[3])

    l, w, h = bbox3d[4:7]

    x_corners = [l / 2, l / 2, -l / 2, -l / 2, l / 2, l / 2, -l / 2, -l / 2]
    y_corners = [0, 0, 0, 0, -h, -h, -h, -h]
    z_corners = [w / 2, -w / 2, -w / 2, w / 2, w / 2, -w / 2, -w / 2, w / 2]

    corners_3d = np.dot(rot_matrix, np.vstack([x_corners, y_corners, z_corners]))
    corners_3d[0, :] = corners_3d[0, :] + bbox3d[0]
    corners_3d[1, :] = corners_3d[1, :] + bbox3d[1]
    corners_3d[2, :] = corners_3d[2, :] + bbox3d[2]

    return np.transpose(corners_3d)



from opendr.perception.object_tracking_3d.ab3dmot.object_tracking_3d_ab3dmot_learner import (
    ObjectTracking3DAb3dmotLearner
)
from opendr.perception.object_tracking_3d.datasets.kitti_tracking import KittiTrackingDatasetIterator

__all__ = ['ObjectTracking3DAb3dmotLearner', 'KittiTrackingDatasetIterator']



# Copyright 2020-2023 OpenDR European Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
from opendr.engine.target import BoundingBox3DList, TrackingAnnotation3DList
from scipy.optimize import linear_sum_assignment
from opendr.perception.object_tracking_3d.ab3dmot.algorithm.core import convert_3dbox_to_8corner, iou3D
from opendr.perception.object_tracking_3d.ab3dmot.algorithm.kalman_tracker_3d import KalmanTracker3D


class AB3DMOT():
    def __init__(
        self, max_staleness=2, min_updates=3, frame=0,
        state_dimensions=10,  # x, y, z, rotation_y, l, w, h, speed_x, speed_z, angular_speed
        measurement_dimensions=7,  # x, y, z, rotation_y, l, w, h
        state_transition_matrix=None,
        measurement_function_matrix=None,
        covariance_matrix=None,
        process_uncertainty_matrix=None,
        iou_threshold=0.01,
    ):
        super().__init__()

        self.max_staleness = max_staleness
        self.min_updates = min_updates
        self.frame = frame - 1
        self.starting_frame = frame - 1
        self.tracklets = []
        self.last_tracklet_id = 0
        self.iou_threshold = iou_threshold

        self.state_dimensions = state_dimensions
        self.measurement_dimensions = measurement_dimensions
        self.state_transition_matrix = state_transition_matrix
        self.measurement_function_matrix = measurement_function_matrix
        self.covariance_matrix = covariance_matrix
        self.process_uncertainty_matrix = process_uncertainty_matrix

    def update(self, detections: BoundingBox3DList):

        self.frame += 1

        if len(detections) > 0:

            predictions = np.zeros([len(self.tracklets), self.measurement_dimensions])

            for i, tracklet in enumerate(self.tracklets):
                box = tracklet.predict().reshape(-1)[:self.measurement_dimensions]
                predictions[i] = [*box]

            detection_corners = [
                convert_3dbox_to_8corner(np.array([*box.location, box.rotation_y, *box.dimensions]))
                for box in detections.boxes
            ]

            if len(predictions) > 0:
                prediction_corners = [
                    convert_3dbox_to_8corner(p)
                    for p in predictions
                ]
            else:
                prediction_corners = np.zeros((0, 8, 3))

            (
                matched_pairs,
                unmatched_detections,
                unmatched_predictions
            ) = associate(detection_corners, prediction_corners, self.iou_threshold)

            for d, p in matched_pairs:
                self.tracklets[p].update(detections[d], self.frame)

            for d in unmatched_detections:
                self.last_tracklet_id += 1
                tracklet = KalmanTracker3D(
                    detections[d], self.last_tracklet_id, self.frame,
                    self.state_dimensions, self.measurement_dimensions,
                    self.state_transition_matrix, self.measurement_function_matrix,
                    self.covariance_matrix, self.process_uncertainty_matrix
                )
                self.tracklets.append(tracklet)

        old_tracklets = self.tracklets
        self.tracklets = []
        tracked_boxes = []

        for tracklet in old_tracklets:

            if tracklet.staleness(self.frame) < self.max_staleness:
                self.tracklets.append(tracklet)

                if self.frame <= self.min_updates or tracklet.updates >= self.min_updates:
                    tracked_boxes.append(tracklet.tracking_bounding_box_3d(self.frame))

        result = TrackingAnnotation3DList(tracked_boxes)
        return result

    def reset(self):
        self.frame = self.starting_frame
        self.tracklets = []
        self.last_tracklet_id = 0


def associate(detection_corners, prediction_corners, iou_threshold):

    iou_matrix = np.zeros((len(detection_corners), len(prediction_corners)), dtype=np.float32)
    for d, det in enumerate(detection_corners):
        for t, trk in enumerate(prediction_corners):
            iou_matrix[d, t] = iou3D(det, trk)[0]

    detection_match_ids, prediction_match_ids = linear_sum_assignment(-iou_matrix)
    unmatched_detections = []
    unmatched_predictions = []

    for i in range(len(detection_corners)):
        if i not in detection_match_ids:
            unmatched_detections.append(i)

    for i in range(len(prediction_corners)):
        if i not in detection_match_ids:
            unmatched_predictions.append(i)

    matched_pairs = []

    for i in range(len(detection_match_ids)):
        detection_id = detection_match_ids[i]
        prediction_id = prediction_match_ids[i]

        if iou_matrix[detection_id, prediction_id] < iou_threshold:
            unmatched_detections.append(detection_id)
            unmatched_predictions.append(prediction_id)
        else:
            matched_pairs.append([detection_id, prediction_id])

    if len(matched_pairs) <= 0:
        matched_pairs = np.zeros((0, 2), dtype=np.int32)
    else:
        matched_pairs = np.array(matched_pairs, dtype=np.int32)

    return matched_pairs, unmatched_detections, unmatched_predictions



# Copyright 2020-2023 OpenDR European Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from opendr.engine.learners import Learner


class ExampleLearner(Learner):
    # 1. The default values in constructor arguments can be set according to the algorithm.
    # 2. Some of the shared parameters, e.g. lr_schedule, backbone, etc., can be skipped here if not needed
    #    by the implementation.
    # 3. TODO Make sure the naming of the arguments is the same as the parent class arguments to keep it consistent
    #     for the end user.
    def __init__(self, lr=0.001, iters=10, batch_size=64, optimizer='sgd', lr_schedule='',
                 backbone='default', network_head='', checkpoint_after_iter=0, checkpoint_load_iter=0,
                 temp_path='', device='cuda', threshold=0.0, scale=1.0,
                 custom_param_1="default val", custom_param_2="default val"):
        # Pass the shared parameters on super's constructor so they can get initialized as class attributes
        super(ExampleLearner, self).__init__(lr=lr, iters=iters, batch_size=batch_size, optimizer=optimizer,
                                             lr_schedule=lr_schedule, backbone=backbone, network_head=network_head,
                                             checkpoint_after_iter=checkpoint_after_iter,
                                             checkpoint_load_iter=checkpoint_load_iter, temp_path=temp_path,
                                             device=device, threshold=threshold, scale=scale)

        # Define the implementation specific parameters
        # TODO Make sure to do appropriate typechecks and provide valid default values for all custom parameters used.
        self.custom_param_1 = custom_param_1
        self.custom_param_2 = custom_param_2

    # All methods below are dummy implementations of the abstract methods that are inherited.
    def save(self, path):
        pass

    def load(self, path):
        pass

    def optimize(self, params):
        pass

    def reset(self):
        pass

    def fit(self, dataset, val_dataset=None, logging_path='', silent=False, verbose=False):
        # The fit method's signature requires all of the above arguments to be present. The algorithm-specific
        # default values can be set here, e.g. set the default value of verbose to True
        pass

    def eval(self, dataset):
        pass

    def infer(self, batch, tracked_bounding_boxes=None):
        # In this infer dummy implementation, a custom argument is added as optional, so as not to change the basic
        # signature of the abstract method.
        # TODO The implementation must make sure it throws an appropriate error if the custom argument is needed and
        #  not provided (None).
        pass


# Overriding shared param through constructor, passing a value into a custom parameter
exampleLearner = ExampleLearner(lr=100.0, custom_param_1="custom value")
print("lr overridden through constructor   :", exampleLearner.lr)
print("batch_size default                  :", exampleLearner.batch_size)
# Overriding shared param after creation
exampleLearner.batch_size = 999
print("batch_size overridden after creation:", exampleLearner.batch_size)
# Printing custom parameters values
print("custom_param_1 overridden through constructor:", exampleLearner.custom_param_1)
print("custom_param_2                               :", exampleLearner.custom_param_2)



# Copyright 2020-2023 OpenDR European Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class Logger:

    LOG_WHEN_SILENT = 3
    LOG_WHEN_NORMAL = 2
    LOG_WHEN_VERBOSE = 1

    def __init__(self, silent, verbose, logfile_path=None):
        super().__init__()

        self.silent = silent
        self.verbose = verbose
        self.logfile = None

        if logfile_path is not None:
            self.logfile = open(logfile_path, "a")

    def log(self, level, *entries, **kwargs):

        if self.silent:
            if level >= Logger.LOG_WHEN_SILENT:
                print(*entries, **kwargs)

                if self.logfile is not None:
                    print(*entries, **kwargs, file=self.logfile)
        elif self.verbose:
            if level >= Logger.LOG_WHEN_VERBOSE:
                print(*entries, **kwargs)

                if self.logfile is not None:
                    print(*entries, **kwargs, file=self.logfile)
        else:
            if level >= Logger.LOG_WHEN_NORMAL:
                print(*entries, **kwargs)

                if self.logfile is not None:
                    print(*entries, **kwargs, file=self.logfile)

    def close(self):
        if self.logfile is not None:
            self.logfile.close()



# Copyright 2020-2023 OpenDR European Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# The code is based on KITTI tracking dev kit code (http://www.cvlibs.net/datasets/kitti/eval_tracking.php)

import copy
import math
import numpy as np
from typing import List
from munkres import Munkres
from collections import defaultdict
from opendr.engine.target import TrackingAnnotation3DList

from opendr.perception.object_detection_3d.voxel_object_detection_3d.second_detector.core.box_np_ops import (
    center_to_corner_box3d,
)
from opendr.perception.object_tracking_3d.ab3dmot.logger import Logger
from numba.cuda.cudadrv.error import CudaSupportError

try:
    from opendr.perception.object_detection_3d.voxel_object_detection_3d.\
        second_detector.core.non_max_suppression.nms_gpu import (
            rotate_iou_gpu_eval as iou3D
        )
except (CudaSupportError, ValueError):

    def iou3D(boxes, qboxes, criterion=-1):
        return np.ones((boxes.shape[0], qboxes.shape[0]))


class TrackingData:
    """
        Utility class to load data.
    """

    def __init__(
        self,
        frame=-1,
        obj_type="unset",
        truncation=-1,
        occlusion=-1,
        obs_angle=-10,
        x1=-1,
        y1=-1,
        x2=-1,
        y2=-1,
        w=-1,
        h=-1,
        length=-1,
        X=-1000,
        Y=-1000,
        Z=-1000,
        yaw=-10,
        score=-1000,
        track_id=-1,
    ):
        """
            Constructor, initializes the object given the parameters.
        """

        # init object data
        self.frame = frame
        self.track_id = track_id
        self.obj_type = obj_type
        self.truncation = truncation
        self.occlusion = occlusion
        self.obs_angle = obs_angle
        self.x1 = x1
        self.y1 = y1
        self.x2 = x2
        self.y2 = y2
        self.w = w
        self.h = h
        self.length = length
        self.X = X
        self.Y = Y
        self.Z = Z
        self.yaw = yaw
        self.score = score
        self.ignored = False
        self.valid = False
        self.tracker = -1

    def __str__(self):
        """
            Print read data.
        """

        attrs = vars(self)
        return "\n".join("%s: %s" % item for item in attrs.items())


class TrackingEvaluator(object):
    """ tracking statistics (CLEAR MOT, id-switches, fragments, ML/PT/MT, precision/recall)
             MOTA	- Multi-object tracking accuracy in [0,100]
             MOTP	- Multi-object tracking precision in [0,100] (3D) / [td,100] (2D)
             MOTAL	- Multi-object tracking accuracy in [0,100] with log10(id-switches)

             id-switches - number of id switches
             fragments   - number of fragmentations

             MT, PT, ML	- number of mostly tracked, partially tracked and mostly lost trajectories

             recall	        - recall = percentage of detected targets
             precision	    - precision = percentage of correctly detected targets
             FAR		    - number of false alarms per frame
             falsepositives - number of false positives (FP)
             missed         - number of missed targets (FN)
    """

    def __init__(
        self,
        cls,
        n_sequences,
        n_frames,
        min_overlap=0.5,
        max_truncation=0,
        min_height=25,
        max_occlusion=2,
        log=print,
    ):

        self.n_sequences = n_sequences
        self.n_frames = n_frames
        self.log = log

        # class to evaluate, i.e. pedestrian or car
        self.cls = cls

        # statistics and numbers for evaluation
        self.n_gt = 0  # number of ground truth detections minus ignored false negatives and true positives
        self.n_igt = 0  # number of ignored ground truth detections
        self.n_gts = (
            []
        )  # number of ground truth detections minus ignored false negatives and true positives PER SEQUENCE
        self.n_igts = (
            []
        )  # number of ground ignored truth detections PER SEQUENCE
        self.n_gt_trajectories = 0
        self.n_gt_seq = []
        self.n_tr = (
            0  # number of tracker detections minus ignored tracker detections
        )
        self.n_trs = (
            []
        )  # number of tracker detections minus ignored tracker detections PER SEQUENCE
        self.n_itr = 0  # number of ignored tracker detections
        self.n_itrs = []  # number of ignored tracker detections PER SEQUENCE
        self.n_igttr = 0  # number of ignored ground truth detections
        # where the corresponding associated tracker detection is also ignored
        self.n_tr_trajectories = 0
        self.n_tr_seq = []
        self.MOTA = 0
        self.MOTP = 0
        self.MOTAL = 0
        self.MODA = 0
        self.MODP = 0
        self.MODP_t = []
        self.recall = 0
        self.precision = 0
        self.F1 = 0
        self.FAR = 0
        self.total_cost = 0
        self.itp = 0  # number of ignored true positives
        self.itps = []  # number of ignored true positives PER SEQUENCE
        self.tp = (
            0  # number of true positives including ignored true positives!
        )
        self.tps = (
            []
        )  # number of true positives including ignored true positives PER SEQUENCE
        self.fn = (
            0  # number of false negatives WITHOUT ignored false negatives
        )
        self.fns = (
            []
        )  # number of false negatives WITHOUT ignored false negatives PER SEQUENCE
        self.ifn = 0  # number of ignored false negatives
        self.ifns = []  # number of ignored false negatives PER SEQUENCE
        self.fp = 0  # number of false positives
        # a bit tricky, the number of ignored false negatives and ignored true positives
        # is subtracted, but if both tracker detection and ground truth detection
        # are ignored this number is added again to avoid double counting
        self.fps = []  # above PER SEQUENCE
        self.mme = 0
        self.fragments = 0
        self.id_switches = 0
        self.MT = 0
        self.PT = 0
        self.ML = 0

        self.min_overlap = (
            min_overlap  # minimum bounding box overlap for 3rd party metrics
        )
        self.max_truncation = (
            max_truncation  # maximum truncation of an object for evaluation
        )
        self.max_occlusion = (
            max_occlusion  # maximum occlusion of an object for evaluation
        )
        self.min_height = (
            min_height  # minimum height of an object for evaluation
        )
        self.n_sample_points = 500

        # this should be enough to hold all groundtruth trajectories
        # is expanded if necessary and reduced in any case
        self.gt_trajectories = [[] for x in range(self.n_sequences)]
        self.ign_trajectories = [[] for x in range(self.n_sequences)]

    def load_ground_truths(self, ground_truths: List[TrackingAnnotation3DList]):
        """
            Helper function to load ground truth.
        """

        try:
            self._load_data(
                ground_truths, cls=self.cls, loading_groundtruth=True
            )
        except IOError:
            return False
        return True

    def load_predictions(self, predictions: List[TrackingAnnotation3DList]):
        """
            Helper function to load tracker data.
        """

        try:
            if not self._load_data(
                predictions, cls=self.cls, loading_groundtruth=False
            ):
                return False
        except IOError:
            return False
        return True

    def _load_data(
        self, input_sequences: List[TrackingAnnotation3DList], cls, min_score=-1000, loading_groundtruth=False
    ):
        """
            Generic loader for ground truth and tracking data.
            Use loadGroundtruth() or loadTracker() to load this data.
            Loads detections in KITTI format from textfiles.
        """

        # classes that should be loaded (ignored neighboring classes)
        if "car" in cls.lower():
            classes = ["car", "van"]
        elif "pedestrian" in cls.lower():
            classes = ["pedestrian", "person_sitting"]
        else:
            classes = [cls.lower()]

        if loading_groundtruth:
            classes += ["dontcare"]

        # construct objectDetections object to hold detection data
        t_data = TrackingData()
        eval_2d = True
        eval_3d = True

        seq_data = []
        n_trajectories = 0
        n_trajectories_seq = []
        for seq, input_seq_data in enumerate(input_sequences):
            ids = []
            n_in_seq = 0
            id_frame_cache = []

            input_seq_boxes = []

            for boxList in input_seq_data:
                input_seq_boxes += boxList.boxes

            # f_data = [[] for x in range(input_seq_boxes[-1].frame + 1)]
            f_data = [[] for x in range(len(input_seq_data))]

            for trackingAnnotation3D in input_seq_boxes:
                # KITTI tracking benchmark data format:
                # (frame,tracklet_id,objectType,truncation,occlusion,alpha,x1,y1,x2,y2,h,w,l,X,Y,Z,ry)

                if not any([s for s in classes if s == trackingAnnotation3D.name.lower()]):
                    continue
                # get fields from table
                t_data.frame = int(trackingAnnotation3D.frame)
                t_data.track_id = int(trackingAnnotation3D.id)
                t_data.obj_type = trackingAnnotation3D.name.lower()  # object type [car, pedestrian, cyclist, ...]
                t_data.truncation = int(
                    trackingAnnotation3D.truncated
                )  # truncation [-1,0,1,2]
                t_data.occlusion = int(
                    trackingAnnotation3D.occluded
                )  # occlusion  [-1,0,1,2]
                t_data.obs_angle = float(trackingAnnotation3D.alpha)  # observation angle [rad]
                t_data.x1 = float(trackingAnnotation3D.bbox2d[0])  # left   [px]
                t_data.y1 = float(trackingAnnotation3D.bbox2d[1])  # top    [px]
                t_data.x2 = float(trackingAnnotation3D.bbox2d[2])  # right  [px]
                t_data.y2 = float(trackingAnnotation3D.bbox2d[3])  # bottom [px]
                t_data.h = float(trackingAnnotation3D.dimensions[0])  # height [m]
                t_data.w = float(trackingAnnotation3D.dimensions[1])  # width  [m]
                t_data.length = float(trackingAnnotation3D.dimensions[2])  # length [m]
                t_data.X = float(trackingAnnotation3D.location[0])  # X [m]
                t_data.Y = float(trackingAnnotation3D.location[1])  # Y [m]
                t_data.Z = float(trackingAnnotation3D.location[2])  # Z [m]
                t_data.yaw = float(trackingAnnotation3D.rotation_y)  # yaw angle [rad]
                t_data.score = float(trackingAnnotation3D.confidence)

                # do not consider objects marked as invalid
                if t_data.track_id == -1 and t_data.obj_type != "dontcare":
                    continue

                idx = t_data.frame
                # check if length for frame data is sufficient
                if idx >= len(f_data):
                    raise ValueError("Frame " + str(idx) + " is out of range")

                id_frame = (t_data.frame, t_data.track_id)
                if id_frame in id_frame_cache and not loading_groundtruth:
                    raise ValueError(
                        "track ids are not unique for sequence %d: frame %d"
                        % (seq, t_data.frame) + (
                            " track id %d occured at least twice for this frame"
                            % t_data.track_id
                        )
                    )
                id_frame_cache.append(id_frame)
                f_data[t_data.frame].append(copy.copy(t_data))

                if (
                    t_data.track_id not in ids and
                    t_data.obj_type != "dontcare"
                ):
                    ids.append(t_data.track_id)
                    n_trajectories += 1
                    n_in_seq += 1

                # check if uploaded data provides information for 2D and 3D evaluation
                if (
                    not loading_groundtruth and
                    eval_2d is True and
                    (
                        t_data.x1 == -1 or
                        t_data.x2 == -1 or
                        t_data.y1 == -1 or
                        t_data.y2 == -1
                    )
                ):
                    eval_2d = False
                if (
                    not loading_groundtruth and
                    eval_3d is True and
                    (
                        t_data.X == -1000 or
                        t_data.Y == -1000 or
                        t_data.Z == -1000
                    )
                ):
                    eval_3d = False

            # only add existing frames
            n_trajectories_seq.append(n_in_seq)
            seq_data.append(f_data)

        if not loading_groundtruth:
            self.tracker = seq_data
            self.n_tr_trajectories = n_trajectories
            self.eval_2d = eval_2d
            self.eval_3d = eval_3d
            self.n_tr_seq = n_trajectories_seq
            if self.n_tr_trajectories == 0:
                return False
        else:
            # split ground truth and DontCare areas
            self.dcareas = []
            self.groundtruth = []
            for seq_idx in range(len(seq_data)):
                seq_gt = seq_data[seq_idx]
                s_g, s_dc = [], []
                for f in range(len(seq_gt)):
                    all_gt = seq_gt[f]
                    g, dc = [], []
                    for gg in all_gt:
                        if gg.obj_type == "dontcare":
                            dc.append(gg)
                        else:
                            g.append(gg)
                    s_g.append(g)
                    s_dc.append(dc)
                self.dcareas.append(s_dc)
                self.groundtruth.append(s_g)
            self.n_gt_seq = n_trajectories_seq
            self.n_gt_trajectories = n_trajectories
        return True

    def box_overlap_3d(self, a: TrackingData, b: TrackingData, criterion="union"):
        """
            boxoverlap computes 3d intersection over union for bbox a and b in KITTI format.
            If the criterion is 'union', overlap = (a inter b) / a union b).
            If the criterion is 'a', overlap = (a inter b) / a, where b should be a dontcare area.
        """

        result = iou3D(
            center_to_corner_box3d(
                np.array([a.X, a.Y, a.Z]).reshape(1, -1),
                np.array([a.h, a.w, a.length]).reshape(1, -1),
                np.array([a.yaw])
            ),
            center_to_corner_box3d(
                np.array([b.X, b.Y, b.Z]).reshape(1, -1),
                np.array([b.h, b.w, b.length]).reshape(1, -1),
                np.array([b.yaw])
            )
        )

        return np.sum(result)

    def compute(self):
        """
            Computes the metrics defined in
                - Stiefelhagen 2008: Evaluating Multiple Object Tracking Performance: The CLEAR MOT Metrics
                  MOTA, MOTAL, MOTP
                - Nevatia 2008: Global Data Association for Multi-Object Tracking Using Network Flows
                  MT/PT/ML
        """

        # construct Munkres object for Hungarian Method association
        hm = Munkres()
        max_cost = 1e9

        # go through all frames and associate ground truth and tracker results
        # groundtruth and tracker contain lists for every single frame containing lists of KITTI format detections
        fr, ids = 0, 0
        for seq_idx in range(len(self.groundtruth)):
            self.log(
                Logger.LOG_WHEN_NORMAL, "Computing metrics " + self.cls +
                "[" + str(seq_idx + 1) + "/" + str(len(self.groundtruth)) + "]", end='\r'
            )
            seq_gt = self.groundtruth[seq_idx]
            seq_dc = self.dcareas[seq_idx]  # don't care areas
            seq_tracker = self.tracker[seq_idx]
            seq_trajectories = defaultdict(list)
            seq_ignored = defaultdict(list)

            # statistics over the current sequence, check the corresponding
            # variable comments in __init__ to get their meaning
            seqtp = 0
            seqitp = 0
            seqfn = 0
            seqifn = 0
            seqfp = 0
            seqigt = 0
            seqitr = 0

            last_ids = [[], []]

            n_gts = 0
            n_trs = 0

            for f in range(len(seq_gt)):

                g = seq_gt[f]
                dc = seq_dc[f]

                t = seq_tracker[f]
                # counting total number of ground truth and tracker objects
                self.n_gt += len(g)
                self.n_tr += len(t)

                n_gts += len(g)
                n_trs += len(t)

                # use hungarian method to associate, using boxoverlap 0..1 as cost
                # build cost matrix
                cost_matrix = []
                this_ids = [[], []]
                for gg in g:
                    # save current ids
                    this_ids[0].append(gg.track_id)
                    this_ids[1].append(-1)
                    gg.tracker = -1
                    gg.id_switch = 0
                    gg.fragmentation = 0
                    cost_row = []
                    for tt in t:
                        # overlap == 1 is cost ==0
                        c = 1 - self.box_overlap_3d(gg, tt)
                        # gating for boxoverlap
                        if c <= self.min_overlap:
                            cost_row.append(c)
                        else:
                            cost_row.append(max_cost)  # = 1e9
                    cost_matrix.append(cost_row)
                    # all ground truth trajectories are initially not associated
                    # extend groundtruth trajectories lists (merge lists)
                    seq_trajectories[gg.track_id].append(-1)
                    seq_ignored[gg.track_id].append(False)

                if len(g) == 0:
                    cost_matrix = [[]]
                # associate
                association_matrix = hm.compute(cost_matrix)

                # tmp variables for sanity checks and MODP computation
                tmptp = 0
                tmpfp = 0
                tmpfn = 0
                tmpc = (
                    0  # this will sum up the overlaps for all true positives
                )
                tmpcs = [0] * len(
                    g
                )  # this will save the overlaps for all true positives
                # the reason is that some true positives might be ignored
                # later such that the corrsponding overlaps can
                # be subtracted from tmpc for MODP computation

                # mapping for tracker ids and ground truth ids
                for row, col in association_matrix:
                    # apply gating on boxoverlap
                    c = cost_matrix[row][col]
                    if c < max_cost:
                        g[row].tracker = t[col].track_id
                        this_ids[1][row] = t[col].track_id
                        t[col].valid = True
                        g[row].distance = c
                        self.total_cost += 1 - c
                        tmpc += 1 - c
                        tmpcs[row] = 1 - c
                        seq_trajectories[g[row].track_id][-1] = t[col].track_id

                        # true positives are only valid associations
                        self.tp += 1
                        tmptp += 1
                    else:
                        g[row].tracker = -1
                        self.fn += 1
                        tmpfn += 1

                # associate tracker and DontCare areas
                # ignore tracker in neighboring classes
                nignoredtracker = 0  # number of ignored tracker detections
                ignoredtrackers = dict()  # will associate the track_id with -1
                # if it is not ignored and 1 if it is
                # ignored;
                # this is used to avoid double counting ignored
                # cases, see the next loop

                for tt in t:
                    ignoredtrackers[tt.track_id] = -1
                    # ignore detection if it belongs to a neighboring class or is
                    # smaller or equal to the minimum height

                    tt_height = abs(tt.y1 - tt.y2)
                    if (
                        (self.cls == "car" and tt.obj_type == "van") or
                        (
                            self.cls == "pedestrian" and
                            tt.obj_type == "person_sitting"
                        ) or
                        tt_height <= self.min_height
                    ) and not tt.valid:
                        nignoredtracker += 1
                        tt.ignored = True
                        ignoredtrackers[tt.track_id] = 1
                        continue
                    for d in dc:
                        overlap = self.box_overlap_3d(tt, d, "a")
                        if overlap > 0.5 and not tt.valid:
                            tt.ignored = True
                            nignoredtracker += 1
                            ignoredtrackers[tt.track_id] = 1
                            break

                # check for ignored FN/TP (truncation or neighboring object class)
                ignoredfn = 0  # the number of ignored false negatives
                nignoredtp = 0  # the number of ignored true positives
                nignoredpairs = (
                    0  # the number of ignored pairs, i.e. a true positive
                )
                # which is ignored but where the associated tracker
                # detection has already been ignored

                gi = 0
                for gg in g:
                    if gg.tracker < 0:
                        if (
                            gg.occlusion > self.max_occlusion or
                            gg.truncation > self.max_truncation or
                            (self.cls == "car" and gg.obj_type == "van") or
                            (
                                self.cls == "pedestrian" and
                                gg.obj_type == "person_sitting"
                            )
                        ):
                            seq_ignored[gg.track_id][-1] = True
                            gg.ignored = True
                            ignoredfn += 1

                    elif gg.tracker >= 0:
                        if (
                            gg.occlusion > self.max_occlusion or
                            gg.truncation > self.max_truncation or
                            (self.cls == "car" and gg.obj_type == "van") or
                            (
                                self.cls == "pedestrian" and
                                gg.obj_type == "person_sitting"
                            )
                        ):

                            seq_ignored[gg.track_id][-1] = True
                            gg.ignored = True
                            nignoredtp += 1

                            # if the associated tracker detection is already ignored,
                            # we want to avoid double counting ignored detections
                            if ignoredtrackers[gg.tracker] > 0:
                                nignoredpairs += 1

                            # for computing MODP, the overlaps from ignored detections
                            # are subtracted
                            tmpc -= tmpcs[gi]
                    gi += 1

                # the below might be confusion, check the comments in __init__
                # to see what the individual statistics represent

                # correct TP by number of ignored TP due to truncation
                # ignored TP are shown as tracked in visualization
                tmptp -= nignoredtp

                # count the number of ignored true positives
                self.itp += nignoredtp

                # adjust the number of ground truth objects considered
                self.n_gt -= ignoredfn + nignoredtp

                # count the number of ignored ground truth objects
                self.n_igt += ignoredfn + nignoredtp

                # count the number of ignored tracker objects
                self.n_itr += nignoredtracker

                # count the number of ignored pairs, i.e. associated tracker and
                # ground truth objects that are both ignored
                self.n_igttr += nignoredpairs

                # false negatives = associated gt bboxes exceding association threshold + non-associated gt bboxes
                #
                tmpfn += len(g) - len(association_matrix) - ignoredfn
                self.fn += len(g) - len(association_matrix) - ignoredfn
                self.ifn += ignoredfn

                # false positives = tracker bboxes - associated tracker bboxes
                # mismatches (mme_t)
                tmpfp += (
                    len(t) -
                    tmptp -
                    nignoredtracker -
                    nignoredtp +
                    nignoredpairs
                )
                self.fp += (
                    len(t) -
                    tmptp -
                    nignoredtracker -
                    nignoredtp +
                    nignoredpairs
                )
                # tmpfp   = len(t) - tmptp - nignoredtp # == len(t) - (tp - ignoredtp) - ignoredtp
                # self.fp += len(t) - tmptp - nignoredtp

                # update sequence data
                seqtp += tmptp
                seqitp += nignoredtp
                seqfp += tmpfp
                seqfn += tmpfn
                seqifn += ignoredfn
                seqigt += ignoredfn + nignoredtp
                seqitr += nignoredtracker

                # sanity checks
                # - the number of true positives minues ignored true positives
                #   should be greater or equal to 0
                # - the number of false negatives should be greater or equal to 0
                # - the number of false positives needs to be greater or equal to 0
                #   otherwise ignored detections might be counted double
                # - the number of counted true positives (plus ignored ones)
                #   and the number of counted false negatives (plus ignored ones)
                #   should match the total number of ground truth objects
                # - the number of counted true positives (plus ignored ones)
                #   and the number of counted false positives
                #   plus the number of ignored tracker detections should
                #   match the total number of tracker detections; note that
                #   nignoredpairs is subtracted here to avoid double counting
                #   of ignored detection sin nignoredtp and nignoredtracker
                if tmptp < 0:
                    raise NameError("Something went wrong! TP is negative")
                if tmpfn < 0:
                    raise NameError("Something went wrong! FN is negative")
                if tmpfp < 0:
                    raise NameError("Something went wrong! FP is negative")
                if tmptp + tmpfn is not len(g) - ignoredfn - nignoredtp:
                    raise NameError(
                        "Something went wrong! nGroundtruth is not TP+FN"
                    )
                if (
                    tmptp +
                    tmpfp +
                    nignoredtp +
                    nignoredtracker -
                    nignoredpairs
                    is not len(t)
                ):
                    raise NameError(
                        "Something went wrong! nTracker is not TP+FP"
                    )

                # check for id switches or fragmentations
                for i, tt in enumerate(this_ids[0]):
                    if tt in last_ids[0]:
                        idx = last_ids[0].index(tt)
                        tid = this_ids[1][i]
                        lid = last_ids[1][idx]
                        if tid != lid and lid != -1 and tid != -1:
                            if g[i].truncation < self.max_truncation:
                                g[i].id_switch = 1
                                ids += 1
                        if tid != lid and lid != -1:
                            if g[i].truncation < self.max_truncation:
                                g[i].fragmentation = 1
                                fr += 1

                # save current index
                last_ids = this_ids
                # compute MOTP_t
                MODP_t = 1
                if tmptp != 0:
                    MODP_t = tmpc / float(tmptp)
                self.MODP_t.append(MODP_t)

            # remove empty lists for current gt trajectories
            self.gt_trajectories[seq_idx] = seq_trajectories
            self.ign_trajectories[seq_idx] = seq_ignored

            # gather statistics for "per sequence" statistics.
            self.n_gts.append(n_gts)
            self.n_trs.append(n_trs)
            self.tps.append(seqtp)
            self.itps.append(seqitp)
            self.fps.append(seqfp)
            self.fns.append(seqfn)
            self.ifns.append(seqifn)
            self.n_igts.append(seqigt)
            self.n_itrs.append(seqitr)

        # compute MT/PT/ML, fragments, idswitches for all groundtruth trajectories
        n_ignored_tr_total = 0
        for seq_idx, (seq_trajectories, seq_ignored) in enumerate(
            zip(self.gt_trajectories, self.ign_trajectories)
        ):
            if len(seq_trajectories) == 0:
                continue
            tmpMT, tmpML, tmpPT, tmpId_switches, tmpFragments = [0] * 5
            n_ignored_tr = 0
            for g, ign_g in zip(
                seq_trajectories.values(), seq_ignored.values()
            ):
                # all frames of this gt trajectory are ignored
                if all(ign_g):
                    n_ignored_tr += 1
                    n_ignored_tr_total += 1
                    continue
                # all frames of this gt trajectory are not assigned to any detections
                if all([this == -1 for this in g]):
                    tmpML += 1
                    self.ML += 1
                    continue
                # compute tracked frames in trajectory
                last_id = g[0]
                # first detection (necessary to be in gt_trajectories) is always tracked
                tracked = 1 if g[0] >= 0 else 0
                lgt = 0 if ign_g[0] else 1
                for f in range(1, len(g)):
                    if ign_g[f]:
                        last_id = -1
                        continue
                    lgt += 1
                    if (
                        last_id != g[f] and
                        last_id != -1 and
                        g[f] != -1 and
                        g[f - 1] != -1
                    ):
                        tmpId_switches += 1
                        self.id_switches += 1
                    if (
                        f < len(g) - 1 and
                        g[f - 1] != g[f] and
                        last_id != -1 and
                        g[f] != -1 and
                        g[f + 1] != -1
                    ):
                        tmpFragments += 1
                        self.fragments += 1
                    if g[f] != -1:
                        tracked += 1
                        last_id = g[f]
                # handle last frame; tracked state is handled in for loop (g[f]!=-1)
                if (
                    len(g) > 1 and
                    g[f - 1] != g[f] and
                    last_id != -1 and
                    g[f] != -1 and
                    not ign_g[f]
                ):
                    tmpFragments += 1
                    self.fragments += 1

                # compute MT/PT/ML
                tracking_ratio = tracked / float(len(g) - sum(ign_g))
                if tracking_ratio > 0.8:
                    tmpMT += 1
                    self.MT += 1
                elif tracking_ratio < 0.2:
                    tmpML += 1
                    self.ML += 1
                else:  # 0.2 <= tracking_ratio <= 0.8
                    tmpPT += 1
                    self.PT += 1

        if (self.n_gt_trajectories - n_ignored_tr_total) == 0:
            self.MT = 0.0
            self.PT = 0.0
            self.ML = 0.0
        else:
            self.MT /= float(self.n_gt_trajectories - n_ignored_tr_total)
            self.PT /= float(self.n_gt_trajectories - n_ignored_tr_total)
            self.ML /= float(self.n_gt_trajectories - n_ignored_tr_total)

        # precision/recall etc.
        if (self.fp + self.tp) == 0 or (self.tp + self.fn) == 0:
            self.recall = 0.0
            self.precision = 0.0
        else:
            self.recall = self.tp / float(self.tp + self.fn)
            self.precision = self.tp / float(self.fp + self.tp)
        if (self.recall + self.precision) == 0:
            self.F1 = 0.0
        else:
            self.F1 = (
                2.0 *
                (self.precision * self.recall) /
                (self.precision + self.recall)
            )
        if sum(self.n_frames) == 0:
            self.FAR = "n/a"
        else:
            self.FAR = self.fp / float(sum(self.n_frames))

        # compute CLEARMOT
        if self.n_gt == 0:
            self.MOTA = -float("inf")
            self.MODA = -float("inf")
        else:
            self.MOTA = 1 - (self.fn + self.fp + self.id_switches) / float(
                self.n_gt
            )
            self.MODA = 1 - (self.fn + self.fp) / float(self.n_gt)
        if self.tp == 0:
            self.MOTP = float("inf")
        else:
            self.MOTP = self.total_cost / float(self.tp)
        if self.n_gt != 0:
            if self.id_switches == 0:
                self.MOTAL = 1 - (
                    self.fn + self.fp + self.id_switches
                ) / float(self.n_gt)
            else:
                self.MOTAL = 1 - (
                    self.fn + self.fp + math.log10(self.id_switches)
                ) / float(self.n_gt)
        else:
            self.MOTAL = -float("inf")
        if sum(self.n_frames) == 0:
            self.MODP = "n/a"
        else:
            self.MODP = sum(self.MODP_t) / float(sum(self.n_frames))
        return True

    def create_summary(self):
        """
            Generate and mail a summary of the results.
            If mailpy.py is present, the summary is instead printed.
        """

        summary = ""

        summary += "tracking evaluation summary".center(80, "=") + "\n"
        summary += (
            self.print_entry(
                "Multiple Object Tracking Accuracy (MOTA)", self.MOTA
            ) +
            "\n"
        )
        summary += (
            self.print_entry(
                "Multiple Object Tracking Precision (MOTP)", self.MOTP
            ) +
            "\n"
        )
        summary += (
            self.print_entry(
                "Multiple Object Tracking Accuracy (MOTAL)", self.MOTAL
            ) +
            "\n"
        )
        summary += (
            self.print_entry(
                "Multiple Object Detection Accuracy (MODA)", self.MODA
            ) +
            "\n"
        )
        summary += (
            self.print_entry(
                "Multiple Object Detection Precision (MODP)", self.MODP
            ) +
            "\n"
        )
        summary += "\n"
        summary += self.print_entry("Recall", self.recall) + "\n"
        summary += self.print_entry("Precision", self.precision) + "\n"
        summary += self.print_entry("F1", self.F1) + "\n"
        summary += self.print_entry("False Alarm Rate", self.FAR) + "\n"
        summary += "\n"
        summary += self.print_entry("Mostly Tracked", self.MT) + "\n"
        summary += self.print_entry("Partly Tracked", self.PT) + "\n"
        summary += self.print_entry("Mostly Lost", self.ML) + "\n"
        summary += "\n"
        summary += self.print_entry("True Positives", self.tp) + "\n"
        summary += self.print_entry("Ignored True Positives", self.itp) + "\n"
        summary += self.print_entry("False Positives", self.fp) + "\n"
        summary += self.print_entry("False Negatives", self.fn) + "\n"
        summary += self.print_entry("Ignored False Negatives", self.ifn) + "\n"
        summary += self.print_entry("Missed Targets", self.fn) + "\n"
        summary += self.print_entry("ID-switches", self.id_switches) + "\n"
        summary += self.print_entry("Fragmentations", self.fragments) + "\n"
        summary += "\n"
        summary += (
            self.print_entry(
                "Ground Truth Objects (Total)", self.n_gt + self.n_igt
            ) +
            "\n"
        )
        summary += (
            self.print_entry("Ignored Ground Truth Objects", self.n_igt) + "\n"
        )
        summary += (
            self.print_entry(
                "Ground Truth Trajectories", self.n_gt_trajectories
            ) +
            "\n"
        )
        summary += "\n"
        summary += self.print_entry("Tracker Objects (Total)", self.n_tr) + "\n"
        summary += (
            self.print_entry("Ignored Tracker Objects", self.n_itr) + "\n"
        )
        summary += (
            self.print_entry("Tracker Trajectories", self.n_tr_trajectories) +
            "\n"
        )
        summary += "=" * 80

        return summary

    def print_entry(self, key, val, width=(70, 10)):
        """
            Pretty print an entry in a table fashion.
        """

        s_out = key.ljust(width[0])
        if type(val) == int:
            s = "%%%dd" % width[1]
            s_out += s % val
        elif type(val) == float:
            s = "%%%df" % (width[1])
            s_out += s % val
        else:
            s_out += ("%s" % val).rjust(width[1])
        return s_out


def evaluate(
    predictions: List[List[TrackingAnnotation3DList]],
    ground_truths: List[List[TrackingAnnotation3DList]],
    log=print
):

    results = {
        "car": None,
        "pedestrian": None,
        "cyclist": None,
    }

    for c in ("car", "pedestrian", "cyclist"):
        evaluator = TrackingEvaluator(
            cls=c, n_sequences=len(ground_truths),
            n_frames=[len(a) for a in ground_truths],
            log=log,
        )
        # load tracker data and check provided classes

        evaluator.load_predictions(predictions)
        evaluator.load_ground_truths(ground_truths)

        if len(evaluator.groundtruth) is not len(evaluator.tracker):
            raise ValueError(
                "The predictions data does not provide results for every sequence."
            )
            return results
        evaluator.compute()

        results[c] = evaluator.create_summary()

    return results



# Copyright 2020-2023 OpenDR European Project
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# KITTI tracking data format:
# (frame,tracklet_id,objectType,truncation,occlusion,alpha,x1,y1,x2,y2,h,w,l,X,Y,Z,ry,[score])
# KITTI detection data format (with frame):
# (frame,objectType,truncation,occlusion,alpha,x1,y1,x2,y2,h,w,l,X,Y,Z,ry,[score])

import os
import time
from zipfile import ZipFile
from urllib.request import urlretrieve
from opendr.engine.constants import OPENDR_SERVER_URL
from opendr.engine.target import (
    BoundingBox3D,
    BoundingBox3DList,
    TrackingAnnotation3D,
    TrackingAnnotation3DList,
)
from opendr.engine.data import PointCloudWithCalibration
from opendr.engine.datasets import DatasetIterator
import numpy as np
from skimage import io

from opendr.perception.object_detection_3d.datasets.kitti import parse_calib


class KittiTrackingDatasetIterator(DatasetIterator):
    def __init__(
        self, inputs_path, ground_truths_path,
        inputs_format="detection"  # detection, tracking
    ):
        super().__init__()

        self.inputs_path = inputs_path
        self.ground_truths_path = ground_truths_path
        self.inputs_format = inputs_format

        self.inputs_files = os.listdir(self.inputs_path)
        self.ground_truths_files = os.listdir(self.ground_truths_path)

        if len(self.inputs_files) != len(self.ground_truths_files):
            raise ValueError(
                "Number of input files and ground_truth files is not identical"
            )

        self.__load_data()

    @staticmethod
    def download(
        url, download_path, inputs_sub_path=".", ground_truths_sub_path=".",
        inputs_format="detection", file_format="zip",
        create_dir=False
    ):

        if file_format == "zip":
            if create_dir:
                os.makedirs(download_path, exist_ok=True)

            print("Downloading KITTI Tracking detections + labels dataset zip file from", url, "to", download_path)

            start_time = 0
            last_print = 0

            def reporthook(count, block_size, total_size):
                nonlocal start_time
                nonlocal last_print
                if count == 0:
                    start_time = time.time()
                    last_print = start_time
                    return

                duration = time.time() - start_time
                progress_size = int(count * block_size)
                speed = int(progress_size / (1024 * duration))
                if time.time() - last_print >= 1:
                    last_print = time.time()
                    print(
                        "\r%d MB, %d KB/s, %d seconds passed" %
                        (progress_size / (1024 * 1024), speed, duration),
                        end=''
                    )

            zip_path = os.path.join(download_path, "dataset.zip")
            urlretrieve(url, zip_path, reporthook=reporthook)
            print()

            print("Extracting KITTI Dataset from zip file")
            with ZipFile(zip_path, 'r') as zip_ref:
                zip_ref.extractall(download_path)

            os.remove(zip_path)

            return KittiTrackingDatasetIterator(
                os.path.join(download_path, inputs_sub_path),
                os.path.join(download_path, ground_truths_sub_path),
                inputs_format
            )

        else:
            raise ValueError("Unsupported file_format: " + file_format)

    @staticmethod
    def download_labels(
        download_path, create_dir=False,
    ):
        return KittiTrackingDatasetIterator.download(
            os.path.join(OPENDR_SERVER_URL, "perception", "object_tracking_3d", "kitti_tracking_labels.zip"),
            download_path,
            "kitti_tracking_labels",
            "kitti_tracking_labels",
            "tracking",
            create_dir=create_dir,
        )

    def __getitem__(self, idx):

        return self.data[idx]

    def __load_data(self):

        data = []

        for input_file, ground_truth_file in zip(
            self.inputs_files,
            self.ground_truths_files
        ):

            input, _ = load_tracking_file(
                os.path.join(self.inputs_path, input_file), self.inputs_format, "detection", remove_dontcare=True
            )
            ground_truth, _ = load_tracking_file(
                os.path.join(self.ground_truths_path, ground_truth_file), "tracking", "tracking"
            )

            data.append((input, ground_truth))

        self.data = data

    def __len__(self):
        return len(self.data)


class LabeledTrackingPointCloudsDatasetIterator(DatasetIterator):
    def __init__(
        self,
        lidar_path,
        label_path,
        calib_path,
        image_path=None,
        labels_format="tracking",  # detection, tracking
        num_point_features=4,
    ):
        super().__init__()

        self.lidar_path = lidar_path
        self.label_path = label_path
        self.calib_path = calib_path
        self.image_path = image_path
        self.num_point_features = num_point_features

        self.lidar_files = sorted(os.listdir(self.lidar_path))
        # self.label_files = sorted(os.listdir(self.label_path))
        # self.calib_files = sorted(os.listdir(self.calib_path))
        self.image_files = (
            sorted(os.listdir(self.image_path))
            if self.image_path is not None
            else None
        )

        self.labels, self.max_id = load_tracking_file(
            self.label_path, "tracking", labels_format,
        )
        self.calib = parse_calib(self.calib_path)

    def __getitem__(self, idx):
        points = np.fromfile(
            os.path.join(self.lidar_path, self.lidar_files[idx]),
            dtype=np.float32,
            count=-1,
        ).reshape([-1, self.num_point_features])
        target = self.labels[idx] if len(self.labels) > idx else TrackingAnnotation3DList([])

        image_shape = (
            None
            if self.image_files is None
            else (
                np.array(
                    io.imread(
                        os.path.join(self.image_path, self.image_files[idx])
                    ).shape[:2],
                    dtype=np.int32,
                )
            )
        )

        result = (
            PointCloudWithCalibration(points, self.calib, image_shape),
            target,
        )

        return result

    def __len__(self):
        return len(self.lidar_files)


def load_tracking_file(
    file_path, format, return_format, remove_dontcare=False
):

    results = {}
    max_frame = -1
    max_id = 0

    with open(file_path) as f:
        lines = [x.strip() for x in f.readlines()]

    for line in lines:
        fields = line.split(" ")
        frame = int(float(fields[0]))

        if format == "tracking":
            if return_format == "tracking":
                box = TrackingAnnotation3D(
                    name=fields[2],
                    truncated=int(float(fields[3])),
                    occluded=int(float(fields[4])),
                    alpha=float(fields[5]),
                    bbox2d=[
                        float(fields[6]),
                        float(fields[7]),
                        float(fields[8]),
                        float(fields[9]),
                    ],
                    dimensions=[
                        float(fields[12]),
                        float(fields[10]),
                        float(fields[11]),
                    ],
                    location=[
                        float(fields[13]),
                        float(fields[14]),
                        float(fields[15]),
                    ],
                    rotation_y=float(fields[16]),
                    score=0 if len(fields) <= 17 else fields[17],
                    frame=int(float(fields[0])),
                    id=int(float(fields[1])),
                )
            elif return_format == "detection":
                box = BoundingBox3D(
                    name=fields[2],
                    truncated=int(float(fields[3])),
                    occluded=int(float(fields[4])),
                    alpha=float(fields[5]),
                    bbox2d=[
                        float(fields[6]),
                        float(fields[7]),
                        float(fields[8]),
                        float(fields[9]),
                    ],
                    dimensions=[
                        float(fields[12]),
                        float(fields[10]),
                        float(fields[11]),
                    ],
                    location=[
                        float(fields[13]),
                        float(fields[14]),
                        float(fields[15]),
                    ],
                    rotation_y=float(fields[16]),
                    score=0 if len(fields) <= 17 else fields[17],
                )
            else:
                raise ValueError(
                    "return_format should be tracking or detection"
                )
        elif format == "detection":
            if return_format == "tracking":
                box = TrackingAnnotation3D(
                    name=fields[1],
                    truncated=int(float(fields[2])),
                    occluded=int(float(fields[3])),
                    alpha=float(fields[4]),
                    bbox2d=[
                        float(fields[5]),
                        float(fields[6]),
                        float(fields[7]),
                        float(fields[8]),
                    ],
                    dimensions=[
                        float(fields[11]),
                        float(fields[9]),
                        float(fields[10]),
                    ],
                    location=[
                        float(fields[12]),
                        float(fields[13]),
                        float(fields[14]),
                    ],
                    rotation_y=float(fields[15]),
                    score=0 if len(fields) <= 15 else fields[16],
                    frame=int(float(fields[0])),
                    id=-1,
                )
            elif return_format == "detection":
                box = BoundingBox3D(
                    name=fields[1],
                    truncated=int(float(fields[2])),
                    occluded=int(float(fields[3])),
                    alpha=float(fields[4]),
                    bbox2d=[
                        float(fields[5]),
                        float(fields[6]),
                        float(fields[7]),
                        float(fields[8]),
                    ],
                    dimensions=[
                        float(fields[11]),
                        float(fields[9]),
                        float(fields[10]),
                    ],
                    location=[
                        float(fields[12]),
                        float(fields[13]),
                        float(fields[14]),
                    ],
                    rotation_y=float(fields[15]),
                    score=0 if len(fields) <= 15 else fields[16],
                )
        else:
            raise ValueError("format should be tracking or detection")

        if frame not in results:
            results[frame] = []
            max_frame = max(max_frame, frame)

        if not (remove_dontcare and box.name == "DontCare"):
            results[frame].append(box)

            if isinstance(box, TrackingAnnotation3D):
                max_id = max(max_id, box.id)

    if return_format == "tracking":

        result = []

        for frame in range(max_frame):
            if frame in results:
                result.append(TrackingAnnotation3DList(results[frame]))
            else:
                result.append(TrackingAnnotation3DList([]))

        return result, max_id
    elif return_format == "detection":
        result = []

        for frame in range(max_frame):
            if frame in results:
                result.append(BoundingBox3DList(results[frame]))
            else:
                result.append(BoundingBox3DList([]))

        return result, max_id
    else:
        raise ValueError("return_format should be tracking or detection")
